import altair as alt
import json
import pandas as pd
import streamlit as st
import streamlit.components.v1 as components

from datetime import timedelta
from pathlib import Path
from streamlit_autorefresh import st_autorefresh
from tclogger import tcdatetime, get_now, get_now_ts

STATUS_COLORS = {
    "create": "#333333",  # gray
    "run": "#ff9944",  # orange
    "done": "#44aa44",  # green
    "idle": "#336677",  # teal
    "loop": "#aa4444",  # red
}
SEP_LABEL_COLOR = "#00ffff"  # cyan
SEP_RULE_COLOR = "#22bbbb"  # cyan
COLOR_DOMAIN = list(STATUS_COLORS.keys())
COLOR_RANGE = list(STATUS_COLORS.values())

TRACK_DAYS = 7
CUTOFF_DAYS = 1
REFRESH_INTERVAL = 15

STATUS_ICONS = {
    "create": "🟢",
    "done": "🟢",
    "idle": "🟢",
    "run": "🟠",
}

PAGE_TITLE = "Actions"
LOGS_DIR = "/home/asimov/repos/bili-scraper/logs"
ACTION_LOG_PATTERN = "action_*.log"


def init_st_page():
    st.set_page_config(
        page_title=PAGE_TITLE,
        page_icon="📡",
        layout="wide",
        initial_sidebar_state="auto",
    )
    st.markdown(
        """
        <style>
            .stAppHeader {
            }
            .stMainBlockContainer{
                padding-top: 0px;
                padding-bottom: 0px;
            }
            h3 {
                font-size: 1.5rem !important;
                padding-top: 0.25rem !important;
                padding-bottom: 0.25rem !important;
            }
            div[data-testid="stElementContainer"]:has(> iframe[height="0"][data-testid="stIFrame"]) {
                display: none !important;
            }
            span[data-baseweb="tag"] {
                background-color: #393939 !important;
            }
        </style>
        """,
        unsafe_allow_html=True,
    )


def init_st_bars():
    st.sidebar.title("Settings")
    log_dir = Path(st.sidebar.text_input("Logs Directory", LOGS_DIR))
    log_files = []
    if log_dir.exists() and log_dir.is_dir():
        log_files = sorted([str(p) for p in log_dir.glob(ACTION_LOG_PATTERN)])
    selected = st.sidebar.multiselect(
        "Select log files",
        options=log_files,
        default=log_files,
        format_func=lambda x: Path(x).name,
    )
    return selected


def inject_title(new_title: str):
    new_title_str = json.dumps(new_title)
    # ts_str is used to trigger page-reload in streamlit
    ts_str = f"<!-- {get_now_ts()} -->"
    set_title_js = (
        f"<script>window.parent.document.title = {new_title_str};\n{ts_str};</script>"
    )
    components.html(set_title_js, height=0)
    log_title_js = f"<script>console.log({new_title_str});\n{ts_str};</script>"
    components.html(log_title_js, height=0)


class ActionMonitor:
    def __init__(self, log_paths):
        self.log_paths = sorted([Path(p) for p in log_paths], reverse=False)
        self.events_by_file = {p.name: [] for p in self.log_paths}

    def get_cutoff_dt(self):
        return get_now() - timedelta(days=CUTOFF_DAYS)

    def get_track_dt(self):
        return get_now() - timedelta(days=TRACK_DAYS)

    def get_x_scale_domain(self):
        cutoff_dt = self.get_cutoff_dt().replace(tzinfo=None)
        now = get_now().replace(tzinfo=None)
        return (cutoff_dt, now)

    def load_events(self):
        for path in self.log_paths:
            evs = []
            if path.exists():
                try:
                    data = json.loads(path.read_text())
                except json.JSONDecodeError:
                    data = []
                for msg in data:
                    dt = tcdatetime.fromisoformat(msg.get("now"))
                    if dt < self.get_track_dt():
                        continue
                    evs.append((dt, msg.get("action")))
                evs.sort(key=lambda x: x[0])
            self.events_by_file[path.name] = evs

    def merge_segs(self, segs: list[dict]) -> list[dict]:
        """
        - If adjacent segs have the same status, merge them into one segment.
        - If timestampe of prev_end is same with next_start,
        then minus 1us of prev_end.
        """
        new_segs = []
        for seg in segs:
            if new_segs and new_segs[-1]["status"] == seg["status"]:
                new_segs[-1]["end"] = seg["end"]
            else:
                new_segs.append(seg)
        for i in range(len(new_segs) - 1):
            prev = new_segs[i]
            next = new_segs[i + 1]
            if prev["end"] == next["start"]:
                prev["end"] = prev["end"] - timedelta(microseconds=1)
        return new_segs

    def get_segs_df(self, events):
        segs = []
        for i in range(len(events) - 1):
            dt, action = events[i]
            next_dt, next_action = events[i + 1]
            if action == "create" and next_action == "run":
                status = "create"
            elif action == "run" and next_action == "done":
                status = "done"
            elif action == "run" and next_action != "done":
                status = "loop"
            elif action == "done" and next_action == "create":
                status = "create"
            elif action == "run":
                status = "run"
            elif action == "done":
                status = "idle"
            else:
                continue
            segs.append({"start": dt, "end": next_dt, "status": status})
        if events:
            dt, action = events[-1]
            if action == "create":
                status = "create"
            elif action == "run":
                status = "run"
            elif action == "done":
                status = "idle"
            else:
                status = None
            if status:
                now = get_now().replace(microsecond=0)
                segs.append({"start": dt, "end": now, "status": status})
        segs = self.merge_segs(segs)
        if not segs:
            return pd.DataFrame(columns=["start", "end", "status"])
        df = pd.DataFrame(segs)
        return df

    def get_seps_df(self, df):
        """get 00:00 in between the min and max datetime in df"""
        min_dt = df["start"].min()
        max_dt = df["end"].max()
        seps = pd.date_range(
            start=min_dt.floor("D"),
            end=max_dt.ceil("D"),
            freq="D",
        )
        seps_df = pd.DataFrame({"sep": seps, "label": seps.strftime("%m-%d")})
        return seps_df

    def format_name(self, name: str):
        name = name.replace("action_", "").replace(".log", "")
        name = "_".join(seg.capitalize() for seg in name.split("_"))
        return name

    def render(self):
        st_autorefresh(interval=REFRESH_INTERVAL * 1000, key="refresh")
        self.load_events()
        # st.title("Actions Monitor")
        # st.write(
        #     f"Timeline over past {TRACK_DAYS} days. Auto-refresh every {REFRESH_INTERVAL}s."
        # )
        status_icons: list[str] = []
        for name, events in self.events_by_file.items():
            header_str = self.format_name(name)
            df = self.get_segs_df(events)
            if df.empty:
                st.subheader(header_str)
                st.info(f"No events in last {TRACK_DAYS} days for {name}.")
                continue
            else:
                last_status = df["status"].iloc[-1]
                status_icon = STATUS_ICONS.get(last_status, "❓")
                status_icons.append(status_icon)
                st.subheader(f"{status_icon} {header_str}")
            bars = (
                alt.Chart(df)
                .mark_bar(size=20)
                .encode(
                    x=alt.X(
                        "start:T",
                        scale=alt.Scale(domain=self.get_x_scale_domain()),
                        axis=alt.Axis(format="%H:%M", labelAngle=0, title=None),
                    ),
                    x2="end:T",
                    y=alt.value(20),
                    color=alt.Color(
                        "status:N",
                        scale=alt.Scale(domain=COLOR_DOMAIN, range=COLOR_RANGE),
                        legend=None,
                    ),
                    tooltip=[
                        alt.Tooltip("start:T", title="Date", format="%Y-%m-%d"),
                        alt.Tooltip(
                            "start:T", title="Start", format="[%m-%d] %H:%M:%S"
                        ),
                        alt.Tooltip("end:T", title="End", format="[%m-%d] %H:%M:%S"),
                        alt.Tooltip("status:N", title="Status"),
                    ],
                )
            )
            chart = bars.properties(height=80, width=800).interactive()
            spec = chart.to_dict()
            spec.setdefault("usermeta", {})["embedOptions"] = {"actions": False}
            st.vega_lite_chart(df, spec, use_container_width=True)
        status_icons_str = "".join(status_icons)
        new_title = f"{status_icons_str} {PAGE_TITLE}"
        inject_title(new_title)


if __name__ == "__main__":
    init_st_page()
    selected = init_st_bars()
    monitor = ActionMonitor(selected)
    monitor.render()

    # streamlit run src/acto/monitor.py --server.address 0.0.0.0



from time import sleep
from tclogger import logger, logstr, brk, get_now_str


class Retrier:
    def __init__(self, max_retries: int = 3, retry_interval: float = 2):
        self.max_retries = max_retries
        self.retry_interval = retry_interval
        self.retry_count = 0

    def log_exception(self, exc_type, exc_val):
        err_mesg = f"× Error {exc_type.__name__}: {exc_val}."
        logger.warn(err_mesg)

    def log_retry_next(self):
        retry_count_str = (
            f"[{logstr.file(self.retry_count)}/{logstr.mesg(self.max_retries)}]"
        )
        now_str = logstr.file(brk(get_now_str()))
        logger.mesg(
            f"{retry_count_str} Retry after {self.retry_interval} seconds ... {now_str}"
        )

    def log_retry_exceed(self):
        logger.warn(f"× Exceed max retries ({self.max_retries}). Exited!")

    def sleep_interval(self):
        sleep(self.retry_interval)

    def is_retry_not_exceeded(self):
        return self.retry_count <= self.max_retries

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def run(self, func, *args, **kwargs):
        while self.is_retry_not_exceeded():
            self.retry_count += 1
            try:
                return func(*args, **kwargs)
            except Exception as e:
                self.log_exception(type(e), e)
                if self.is_retry_not_exceeded():
                    self.log_retry_next()
                    self.sleep_interval()
                else:
                    self.log_retry_exceed()
                    raise e


class SoftRetrier:
    """For funcs do not raise exception, but return False when failed."""

    def __init__(self, max_retries: int = 3, retry_interval: float = 2):
        self.max_retries = max_retries
        self.retry_interval = retry_interval
        self.retry_count = 0

    def is_retry_not_exceeded(self):
        return self.retry_count <= self.max_retries

    def sleep_interval(self):
        sleep(self.retry_interval)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def run(self, func, *args, **kwargs) -> bool:
        while self.is_retry_not_exceeded():
            self.retry_count += 1
            res_bool = func(*args, **kwargs)
            if res_bool is False:
                if self.is_retry_not_exceeded():
                    self.sleep_interval()
                    continue
                else:
                    return False
            else:
                return True



import time

from datetime import datetime
from pathlib import Path
from tclogger import logger, logstr, brk
from tclogger import get_now_str, get_now_ts, str_to_ts, dt_to_str
from tclogger import TCLogbar, add_fills
from typing import Union

from .times import PatternedDatetimeSeeker
from .log import ActionLogger


class Perioder:
    def __init__(
        self,
        patterns: Union[str, dict, list],
        name: str = None,
        log_path: Union[str, Path] = None,
        clock_precision: float = 0.25,
        verbose: bool = True,
    ):
        self.patterns = patterns
        self.log_path = log_path
        self.name = name
        self.clock_precision = clock_precision
        self.verbose = verbose
        self.seeker = PatternedDatetimeSeeker(patterns)
        self.bar = TCLogbar()

    def bind(self, func: callable, desc_func: callable = None):
        self.func = func
        self.desc_func = desc_func
        self.name = self.name or func.__name__
        self.action_logger = ActionLogger(self.log_path or f"{self.func.__name__}.log")

    def log_before_wait(self):
        remain_seconds_str = logstr.file(f"{self.remain_seconds}s")
        remain_dt_str = logstr.file(dt_to_str(int(self.remain_seconds)))
        fill_str = add_fills("", filler="= ")
        create_cli_msg = (
            f"{fill_str}\n"
            f"now: {logstr.file(brk(get_now_str()))}, "
            f"next_run: {logstr.file(brk(self.run_dt_str))}, "
            f"wait_for: {remain_seconds_str} ({remain_dt_str})"
        )
        logger.note(create_cli_msg, verbose=self.verbose)

    def update_desc_and_func(self):
        self.func_strs = []
        self.desc_str = self.func.__name__
        if self.desc_func and callable(self.desc_func):
            self.func_strs, self.desc_str = self.desc_func(self.run_dt_str)

    def action_log_before_wait(self):
        wait_for_str = dt_to_str(int(self.remain_seconds))
        self.create_msg = {
            "name": self.name,
            "now": get_now_str(),
            "type": "period",
            "action": "create",
            "info": {
                "run_at": self.run_dt_str,
                "wait_for": wait_for_str,
                "cmds": self.func_strs,
            },
        }
        self.action_logger.log(self.create_msg)

    def log_wait_progress(self):
        remain_seconds = self.remain_seconds
        total = int(remain_seconds)
        run_dt_ts = str_to_ts(self.run_dt_str)
        self.bar.total = total
        if len(self.desc_str) <= 75:
            self.bar.head = logstr.note(self.desc_str)
        else:
            logger.note(self.desc_str)
        while remain_seconds > 2 * self.clock_precision:
            now_ts = datetime.now().timestamp()
            self.bar.update(
                count=round(total - remain_seconds),
                remain_seconds=round(remain_seconds),
            )
            if now_ts >= run_dt_ts:
                break
            remain_seconds = run_dt_ts - now_ts
            time.sleep(self.clock_precision)
        self.bar.update(count=total, remain_seconds=0, flush=True)
        self.bar.reset(linebreak=True)
        time.sleep(max(run_dt_ts - datetime.now().timestamp(), 0))

    def log_before_func(self):
        single_fill_str = add_fills("", filler="- ")
        logger.mesg(single_fill_str, verbose=self.verbose)

    def action_log_before_func(self):
        self.run_msg = {
            "name": self.name,
            "now": get_now_str(),
            "type": "period",
            "action": "run",
            "info": {
                "run_at": self.run_dt_str,
                "cmds": self.func_strs,
            },
        }
        self.action_logger.log(self.run_msg)

    def log_after_func(self):
        pass

    def action_log_after_func(self):
        elapsed_seconds = get_now_ts() - str_to_ts(self.run_dt_str)
        elapsed_str = dt_to_str(elapsed_seconds)
        done_msg = {
            "name": self.name,
            "now": get_now_str(),
            "type": "period",
            "action": "done",
            "info": {
                "run_at": self.run_dt_str,
                "elapsed": elapsed_str,
            },
        }
        self.action_logger.log(done_msg)

    def run(self):
        for run_dt_str, remain_seconds in self.seeker:
            self.run_dt_str = run_dt_str
            self.remain_seconds = remain_seconds
            self.log_before_wait()
            self.update_desc_and_func()
            self.action_log_before_wait()
            self.log_wait_progress()
            self.log_before_func()
            self.action_log_before_func()
            self.func()
            self.log_after_func()
            self.action_log_after_func()



from .times import PatternedDatetimeSeeker, UnitTimeDistConverter
from .periods import Perioder
from .log import ActionLogger
from .retry import Retrier, SoftRetrier
from .email import Emailer, EmailConfigsType, EmailContentType



import json
import threading

from pathlib import Path
from tclogger import logger, get_now_str
from typing import Literal, Union, TypedDict

DURATION_TYPE = Union[int, float, str]


class PeriodInfoCreate(TypedDict):
    cmds: list[str] = None
    run_at: str = None  # time when run starts
    wait_for: DURATION_TYPE = None  # time to wait for run


class PeriodInfoRun(TypedDict):
    cmds: list[str]
    run_at: str = None  # time when run starts
    waited: DURATION_TYPE = None  # time waited after create


class PeriodInfoDone(TypedDict):
    cmds: list[str] = None
    run_at: str = None  # time when run starts
    elapsed: DURATION_TYPE = None  # time elapsed after run


class PeriodMsgType(TypedDict):
    name: str
    now: str
    type: str = "period"
    action: Literal["create", "run", "done"]
    info: Union[PeriodInfoCreate, PeriodInfoRun, PeriodInfoDone]


class ActionLogger:
    def __init__(self, log_path: Union[str, Path], lock: threading.Lock = None):
        self.log_path = Path(log_path)
        self.lock = lock or threading.Lock()
        self.log_path.parent.mkdir(parents=True, exist_ok=True)

    def log(self, msg: PeriodMsgType):
        try:
            with open(self.log_path, "r") as f:
                json_data = json.load(f)
        except:
            json_data = []
        json_data.append(msg)
        with self.lock:
            with open(self.log_path, "w") as f:
                json.dump(json_data, f, indent=4)



import re

from collections.abc import Generator
from datetime import datetime, timedelta
from operator import add, sub
from itertools import product
from tclogger import logger, logstr, Runtimer, dict_to_str
from tclogger import get_now, get_now_str, str_to_ts, t_to_str, str_to_ts, str_to_t
from typing import Union, Literal

"""
ISO-format str:
    - 'YYYY-mm-dd HH:MM:SS'
    - 'mm-dd HH:MM:SS'
    - 'YYYY-mm-dd'
    - 'mm-dd'
    - 'HH:MM:SS'
    ' 'MM:SS'
"""

FULL_ISO_MASK = "****-**-** **:**:**"
FULL_ISO_ZERO = "0000-00-00 00:00:00"


def is_dt_str_full_iso(dt_str: str) -> bool:
    return len(dt_str) == len(FULL_ISO_MASK)


def mask_dt_str(
    dt_str: str,
    lmask: str = FULL_ISO_MASK,
    rmask: str = FULL_ISO_ZERO,
    start: int = None,
    end: int = None,
) -> str:
    if start is None:
        start = 0
    if end is None:
        end = len(dt_str)
    return lmask[:start] + dt_str + rmask[end:]


def fill_iso(
    dt_str: str,
    lmask: str = FULL_ISO_MASK,
    rmask: str = FULL_ISO_ZERO,
) -> str:
    dt_len = len(dt_str)
    full_len = len(FULL_ISO_MASK)
    if dt_len == full_len:
        return dt_str
    masked_str = re.sub(r"\d", "*", dt_str)
    # slide window from right to left
    for i in range(full_len - dt_len):
        end = full_len - i
        start = end - dt_len
        if FULL_ISO_MASK[start:end] == masked_str:
            return lmask[:start] + dt_str + rmask[end:]
    raise ValueError(f"× Cannot fill ISO format: {dt_str}")


def is_dt_str_valid(dt_str: str) -> bool:
    try:
        str_to_ts(dt_str)
        return True
    except:
        return False


def is_dt_str_later_equal(dt_str1: str, dt_str2: str) -> bool:
    return str_to_ts(dt_str1) >= str_to_ts(dt_str2)


def is_dt_str_later(dt_str1: str, dt_str2: str) -> bool:
    return str_to_ts(dt_str1) > str_to_ts(dt_str2)


def is_dt_str_valid_and_later(dt_str: str, dt_beg: str) -> bool:
    return is_dt_str_valid(dt_str) and is_dt_str_later(dt_str, dt_beg)


def unify_dt_beg_end(
    dt_be: str = None, pos: Literal["beg", "end"] = "beg"
) -> tuple[str, datetime]:
    if pos == "beg":
        bound_str = get_now_str()
    else:
        bound_str = "9999-12-31 23:59:59"
    if dt_be is None:
        dt_be = bound_str
    else:
        dt_be = fill_iso(dt_be, lmask=bound_str, rmask=bound_str)
    t_be = str_to_t(dt_be)
    return dt_be, t_be


def zip_kvs(keys: Union[list[str], str], vals: Union[list[str], str]) -> dict:
    if isinstance(keys, (list, tuple)) and isinstance(vals, (list, tuple)):
        return dict(zip(keys, vals))
    key = keys[0] if isinstance(keys, (list, tuple)) else keys
    val = vals[0] if isinstance(vals, (list, tuple)) else vals
    return {key: val}


def tuplize_vars(v1: Union[tuple, str, int], v2: Union[tuple, str, int]) -> tuple:
    if isinstance(v1, (str, int)):
        v1 = (v1,)
    if isinstance(v2, (str, int)):
        v2 = (v2,)
    return v1 + v2


def delta_seconds(dt_str1: str, dt_str2: str) -> float:
    return (
        datetime.fromisoformat(dt_str2).timestamp()
        - datetime.fromisoformat(dt_str1).timestamp()
    )


def re_match_nn(pattern: str, num: Union[int, str], digits: int = 2) -> bool:
    return re.match(pattern, f"{num:0{digits}d}")


class UnitTimeDistConverter:
    UNIT_SECONDS = {
        "year": {
            "units": ["y", "yr", "year", "years"],
            "seconds": 365 * 24 * 60 * 60,
        },
        "month": {
            "units": ["m", "mon", "month", "months"],
            "seconds": 30 * 24 * 60 * 60,
        },
        "week": {
            "units": ["w", "wk", "week", "weeks"],
            "seconds": 7 * 24 * 60 * 60,
        },
        "day": {
            "units": ["d", "day", "days"],
            "seconds": 24 * 60 * 60,
        },
        "hour": {
            "units": ["h", "hr", "hour", "hours"],
            "seconds": 60 * 60,
        },
        "minute": {
            "units": ["n", "min", "minute", "minutes"],
            "seconds": 60,
        },
        "second": {
            "units": ["s", "sec", "second", "seconds"],
            "seconds": 1,
        },
    }
    RE_NUM_UNIT = r"(\d+)\s*(\S+)"

    def __init__(self, direction: Literal["before", "after"] = "before") -> None:
        self.direction = direction
        self.init_op()

    def init_op(self):
        if self.direction == "before":
            self.op = sub
        else:
            self.op = add

    def get_dt(self, dt_str: str = None) -> datetime:
        if dt_str:
            return datetime.fromisoformat(dt_str)
        else:
            return datetime.now()

    def match_unit_num(self, ut_str: str) -> tuple[str, int, int]:
        match_res = re.match(self.RE_NUM_UNIT, ut_str.strip())
        if match_res:
            num = int(match_res.group(1))
            unit = match_res.group(2).lower()
        else:
            return None, None, None

        matched_unit_name = None
        for unit_name, unit_dict in self.UNIT_SECONDS.items():
            if unit in unit_dict["units"]:
                delta_seconds = num * unit_dict["seconds"]
                matched_unit_name = unit_name

        return matched_unit_name, num, delta_seconds

    def to_datetime(self, ut_str: str, dt_str: str = None) -> datetime:
        dt = self.get_dt(dt_str)
        matched_unit_name, num, delta_seconds = self.match_unit_num(ut_str)
        if matched_unit_name is None:
            return None
        elif matched_unit_name == "year":
            res_dt = datetime(
                self.op(dt.year, num), dt.month, dt.day, dt.hour, dt.minute, dt.second
            )
        elif matched_unit_name == "month":
            res_dt = datetime(
                self.op(dt.year, num // 12),
                (self.op(dt.month, num) - 1) % 12 + 1,
                dt.day,
                dt.hour,
                dt.minute,
                dt.second,
            )
        else:
            res_dt = self.op(dt, timedelta(seconds=delta_seconds))

        return res_dt

    def to_timestamp(self, ut_str: str, dt_str: str = None) -> int:
        res_dt = self.to_datetime(ut_str, dt_str)
        if res_dt:
            return res_dt.timestamp()
        else:
            return None

    def to_dt_str(self, ut_str: str, dt_str: str = None) -> str:
        res_dt = self.to_datetime(ut_str, dt_str)
        if res_dt:
            return t_to_str(res_dt)
        else:
            return None


class YmdhmsPatternConverter:
    LEVELS = ["year", "month", "day", "hour", "minute", "second"]

    def __init__(self) -> None:
        pass

    def ymdhms_pattern_to_re_dict(self, pattern: str) -> dict:
        ymdh_list = re.findall(r"[\d\*]+", pattern)
        yyyy, mm, dd, hh, mi, ss = list(map(lambda x: x.replace("*", "\d"), ymdh_list))
        return {
            "year": yyyy,
            "month": mm,
            "day": dd,
            "hour": hh,
            "minute": mi,
            "second": ss,
        }

    def fill_low_level_as_zero(self, pattern: dict) -> dict:
        pattern = {k: v.replace("*", "\d") for k, v in pattern.items()}
        for level in reversed(self.LEVELS):
            if pattern.get(level) is None:
                pattern[level] = "00"
            else:
                break
        return pattern

    def convert(self, pattern: Union[str, dict, list]) -> dict:
        if isinstance(pattern, str):
            return self.ymdhms_pattern_to_re_dict(pattern)
        if isinstance(pattern, dict):
            return self.fill_low_level_as_zero(pattern)


class PatternedDatetimeSeeker:
    DD_RANGES = {
        "month": range(1, 13),
        "month_r": range(12, 0, -1),
        "day": range(1, 32),
        "day_r": range(31, 0, -1),
        "hour": range(0, 24),
        "hour_r": range(23, -1, -1),
        "minute": range(0, 60),
        "minute_r": range(59, -1, -1),
        "second": range(0, 60),
        "second_r": range(59, -1, -1),
    }
    DD_ATTRS = {
        "year": "yyyy",
        "month": "mm",
        "day": "dd",
        "hour": "hh",
        "minute": "mi",
        "second": "ss",
    }

    def __init__(
        self,
        patterns: Union[str, dict, list[str], list[dict]],
        dt_beg: str = None,
        dt_end: str = None,
    ) -> None:
        self.min_year = None
        self.min_month = None
        self.min_day = None
        self.min_hour = None
        self.min_minute = None
        self.min_second = None
        self.dt_beg = dt_beg
        self.dt_end = dt_end
        self.patterns = patterns
        self.init_dt_beg_end()
        self.init_patterns()

    def init_dt_beg_end(self) -> tuple[str, str]:
        self.dt_beg, self.t_beg = unify_dt_beg_end(self.dt_beg, "beg")
        self.dt_end, self.t_end = unify_dt_beg_end(self.dt_end, "end")
        self.DD_RANGES["year"] = range(self.t_beg.year, 9999)

    def init_patterns(self) -> str:
        converter = YmdhmsPatternConverter()
        self.pattern_dict = converter.convert(self.patterns)

    def iter_dd(self, level: str) -> Generator[str, None, None]:
        level_prefix = level.split("_")[0]
        dd_range = self.DD_RANGES[level]
        pattern = self.pattern_dict.get(level_prefix, None)
        for dd in dd_range:
            if pattern is None or re_match_nn(pattern, dd):
                yield dd

    def iter_dds_product(self, levels: list[str]) -> Generator[tuple[str], None, None]:
        for dds in product(*[self.iter_dd(level) for level in levels]):
            yield dds

    def is_ymdhms_valid_and_later(
        self,
        year: int = None,
        month: int = None,
        day: int = None,
        hour: int = None,
        minute: int = None,
        second: int = None,
        day_end: bool = False,
    ) -> bool:
        yy = f"{year:04d}" if year is not None else self.min_year
        mm = f"{month:02d}" if month is not None else self.min_month
        dd = f"{day:02d}" if day is not None else self.min_day
        hh = f"{hour:02d}" if hour is not None else ("23" if day_end else self.min_hour)
        mi = (
            f"{minute:02d}"
            if minute is not None
            else ("59" if day_end else self.min_minute)
        )
        ss = (
            f"{second:02d}"
            if second is not None
            else ("59" if day_end else self.min_second)
        )
        dt_str = f"{yy}-{mm}-{dd} {hh}:{mi}:{ss}"
        return is_dt_str_valid_and_later(dt_str, self.dt_beg)

    def iter_dds_product_and_check(
        self,
        levels: list[str],
        dd_type: Literal["ymd", "hms"] = "ymd",
        extra_level_kvs: dict = {},
    ) -> Generator[tuple[str], None, None]:
        level_keys = [level.split("_")[0] for level in levels]
        if len(levels) == 1:
            product_iter = self.iter_dd(levels[0])
        else:
            product_iter = self.iter_dds_product(levels)
        for dd in product_iter:
            if self.is_ymdhms_valid_and_later(
                **zip_kvs(level_keys, dd), **extra_level_kvs, day_end=(dd_type == "ymd")
            ):
                yield dd

    def double_iter_dds_product_and_check(
        self, ymd_levels: list[str], hms_levels: list[str]
    ) -> Generator[tuple[str], None, None]:
        ymd_level_keys = [level.split("_")[0] for level in ymd_levels]
        for ymd in self.iter_dds_product_and_check(ymd_levels, "ymd"):
            for hms in self.iter_dds_product_and_check(
                hms_levels, "hms", extra_level_kvs=zip_kvs(ymd_level_keys, ymd)
            ):
                yield tuplize_vars(ymd, hms)

    def calc_min_matched_year(self) -> str:
        for year, *_ in self.double_iter_dds_product_and_check(
            ["year", "month_r", "day_r"], ["hour_r", "minute_r", "second_r"]
        ):
            min_year = f"{year:04d}"
            self.min_year = min_year
            return min_year

    def calc_min_matched_month(self) -> str:
        for month, *_ in self.double_iter_dds_product_and_check(
            ["month", "day_r"], ["hour_r", "minute_r", "second_r"]
        ):
            self.min_month = f"{month:02d}"
            return self.min_month

    def calc_min_matched_day(self) -> str:
        for day, *_ in self.double_iter_dds_product_and_check(
            ["day"], ["hour_r", "minute_r", "second_r"]
        ):
            self.min_day = f"{day:02d}"
            return self.min_day

    def calc_min_matched_hour(self) -> str:
        for hour, *_ in self.double_iter_dds_product_and_check(
            ["hour"], ["minute_r", "second_r"]
        ):
            self.min_hour = f"{hour:02d}"
            return self.min_hour

    def calc_min_matched_minute(self) -> str:
        for minute, *_ in self.double_iter_dds_product_and_check(
            ["minute"], ["second_r"]
        ):
            self.min_minute = f"{minute:02d}"
            return self.min_minute

    def calc_min_matched_second(self) -> str:
        for second in self.iter_dds_product_and_check(["second"], "hms"):
            self.min_second = f"{second:02d}"
            return self.min_second

    def get_min_matched_dt_str(self) -> str:
        self.calc_min_matched_year()
        self.calc_min_matched_month()
        self.calc_min_matched_day()
        self.calc_min_matched_hour()
        self.calc_min_matched_minute()
        self.calc_min_matched_second()
        return f"{self.min_year}-{self.min_month}-{self.min_day} {self.min_hour}:{self.min_minute}:{self.min_second}"

    def __iter__(self) -> Generator[tuple[str, float], None, None]:
        while self.dt_beg <= self.dt_end:
            next_dt_str = self.get_min_matched_dt_str()
            remain_seconds = delta_seconds(self.dt_beg, next_dt_str)
            self.dt_beg = next_dt_str
            yield next_dt_str, remain_seconds


def test_fill_iso():
    logger.note("> test_fill_iso")
    dt_strs = ["00:30", "00:30:", "12-31", "12:00:00", "1 12", "01 **:00:00"]
    with logger.temp_indent(2):
        for dt_str in dt_strs:
            logger.mesg(fill_iso(dt_str))
        now_str = get_now_str()
        for dt_str in dt_strs:
            logger.file(fill_iso(dt_str, lmask=now_str))


def test_unit_time_disk_converter():
    ut_strs = ["1 year", "1 month", "1 week", "1 day", "1 hour", "15 minute"]
    b_converter = UnitTimeDistConverter("before")
    a_converter = UnitTimeDistConverter("after")
    logger.note("> test_unit_time_disk_converter")
    logger.mesg(f"* now: {get_now_str()}")
    res_dict = {}
    for ut_str in ut_strs:
        b_dt_str = b_converter.to_dt_str(ut_str)
        a_dt_str = a_converter.to_dt_str(ut_str)
        res_dict[f"{ut_str} before now"] = b_dt_str
        res_dict[f"{ut_str} after now"] = a_dt_str
    logger.mesg(dict_to_str(res_dict), indent=2)

    base_dt_str = "2025-01-01 00:00:00"
    res_dict = {}
    logger.mesg(f"* base_dt_str: {base_dt_str}")
    for ut_str in ut_strs:
        b_dt_str = b_converter.to_dt_str(ut_str, base_dt_str)
        a_dt_str = a_converter.to_dt_str(ut_str, base_dt_str)
        res_dict[f"{ut_str} before"] = b_dt_str
        res_dict[f"{ut_str} after"] = a_dt_str
    logger.mesg(dict_to_str(res_dict), indent=2)


def test_get_min_matched_dt_str():
    logger.note("> test_get_min_matched_dt_str")
    pattern_answers = [
        ("****-**-** **:**:**", "2025-01-25 02:58:31", "2025-01-25 02:58:32"),
        ("****-**-** 01:**:**", "2025-12-30 00:00:00", "2025-12-30 01:00:00"),
        ("****-**-** 00:**:**", "2025-12-31 23:00:00", "2026-01-01 00:00:00"),
        ("****-*3-** *1:3*:**", "2025-04-30 00:00:00", "2026-03-01 01:30:00"),
        ("****-*3-** *5:**:**", "2022-04-30 00:00:00", "2023-03-01 05:00:00"),
        ("****-*3-** *5:2*:**", "2022-04-30 00:00:00", "2023-03-01 05:20:00"),
        ("****-*2-** 00:00:00", "2024-02-28 00:00:00", "2024-02-29 00:00:00"),
        ("****-*2-** 00:00:00", "2025-02-28 00:00:00", "2025-12-01 00:00:00"),
        ("****-02-** 00:00:00", "2025-02-28 00:00:00", "2026-02-01 00:00:00"),
        ("****-**-*1 00:00:00", "2025-02-28 00:00:00", "2025-03-01 00:00:00"),
        ({"day": "03"}, "2025-02-28 00:00:00", "2025-03-03 00:00:00"),
        ({"hour": "0[34]"}, "2025-02-28 00:00:00", "2025-02-28 03:00:00"),
    ]
    with logger.temp_indent(2):
        for pattern, dt_beg, answer in pattern_answers:
            matcher = PatternedDatetimeSeeker(pattern, dt_beg)
            res = matcher.get_min_matched_dt_str()
            # matcher.dt_beg = res
            # res = matcher.get_min_matched_dt_str()
            if res == answer:
                mark = logstr.success("✓")
                answer_str = logstr.success(answer)
            else:
                mark = logstr.warn("×")
                answer_str = f"{logstr.warn(res)} ≠ {logstr.success(answer)}"
            dt_beg_str = logstr.file(f"(after {dt_beg})")
            logger.mesg(f"{mark} {pattern} {dt_beg_str} → {answer_str}")


if __name__ == "__main__":
    with Runtimer():
        # test_fill_iso()
        # test_get_min_matched_dt_str()
        test_unit_time_disk_converter()

    # python -m acto.times



from tclogger import shell_cmd, logger, get_now_str

from acto.periods import Perioder
from acto.retry import Retrier, SoftRetrier


def foo():
    cmd = 'date +"%T.%N"'
    shell_cmd(cmd, showcmd=False)


def foo_desc_func(x):
    func_strs = ['date +"%T.%N"']
    desc_str = f"foo at {x}"
    return func_strs, desc_str


def test_perioder():
    logger.note("> test_perioder")
    # patterns = "****-**-** **:**:**"
    patterns = {"second": "*[05]"}
    perioder = Perioder(patterns)
    perioder.bind(func=foo, desc_func=foo_desc_func)
    perioder.run()


foo_call_count = 0


def foo_to_retry(ok_until: int = 3):
    global foo_call_count
    foo_call_count += 1

    if foo_call_count < ok_until:
        raise RuntimeError(f"foo_to_retry failed at call: {foo_call_count}")
    else:
        cmd = 'date +"%T.%N"'
        shell_cmd(cmd, showcmd=False)


def test_retrier():
    logger.note("> test_retrier")
    with Retrier(max_retries=3, retry_interval=0.5) as retrier:
        retrier.run(foo_to_retry, ok_until=4)
    logger.mesg("* test_retrier completed")


bar_call_count = 0


def bar_to_soft_retry(ok_until: int = 3):
    global bar_call_count
    bar_call_count += 1
    if bar_call_count < ok_until:
        return False
    else:
        cmd = 'date +"%T.%N"'
        shell_cmd(cmd, showcmd=False)
        print(f"Good at call: {bar_call_count}")
        return True


def test_soft_retrier():
    logger.note("> test_soft_retrier")
    with SoftRetrier(max_retries=3, retry_interval=0.5) as retrier:
        retrier.run(bar_to_soft_retry, ok_until=4)
    logger.mesg("* test_soft_retrier completed")


if __name__ == "__main__":
    # test_perioder()
    # test_retrier()
    test_soft_retrier()

    # python -m acto.tests



import smtplib

from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email import encoders
from tclogger import logger, logstr, dict_to_str, get_now_str, brk, confirm_input
from typing import TypedDict, Optional, Union, Literal
from pathlib import Path


class EmailConfigsType(TypedDict):
    smtp_server: str
    smtp_port: int
    username: str
    password: str


class EmailContentType(TypedDict):
    to: Union[str, list[str]]
    subject: str
    body: str
    cc: Optional[list[str]] = None
    attachments: Optional[list[Union[str, Path]]] = []
    mime_type: Optional[Literal["plain", "html"]] = "html"


class Emailer:
    def __init__(
        self,
        configs: EmailConfigsType,
        connect_at_init: bool = True,
        confirm_before_send: bool = True,
        verbose: bool = True,
    ):
        self.configs = configs
        self.smtp_server = configs["smtp_server"]
        self.smtp_port = configs["smtp_port"]
        self.username = configs["username"]
        self.password = configs["password"]
        self.connect_at_init = connect_at_init
        self.confirm_before_send = confirm_before_send
        self.verbose = verbose
        logger.enter_quiet(not self.verbose)
        self.smtp: Optional[smtplib.SMTP] = None
        if connect_at_init:
            self.connect()

    def connect(self) -> smtplib.SMTP:
        logger.note(f"> Connecting SMTP:")
        info_dict = {
            "smtp_server": self.smtp_server,
            "smtp_port": self.smtp_port,
            "username": self.username,
            "now": get_now_str(),
        }
        logger.mesg(dict_to_str(info_dict), indent=2)
        self.smtp = smtplib.SMTP(self.smtp_server, self.smtp_port)
        # context = ssl.create_default_context()
        # self.smtp.starttls(context=context)
        self.smtp.login(self.username, self.password)
        logger.okay(f"+ Connected SMTP: {logstr.file(brk(self.username))}")
        return self.smtp

    def disconnect(self):
        self.smtp.quit()
        logger.okay(f"- Disconnected SMTP: {logstr.file(brk(self.username))}")

    def mail_list_to_str(self, to: Union[str, list[str]]) -> str:
        if not to:
            return to
        if isinstance(to, str):
            return to
        if isinstance(to, (list, tuple)):
            return ", ".join(to)
        raise ValueError("Invalid type for 'to'. Expected str or list[str].")

    def attach_to_mime(self, attach: Union[str, Path]) -> MIMEBase:
        if not Path(attach).exists():
            logger.warn(f"  * Attachment not found: {logstr.file(brk(attach))}")
            return None
        part = MIMEBase("application", "octet-stream")
        with open(attach, "rb") as rf:
            part.set_payload(rf.read())
        encoders.encode_base64(part)
        part.add_header(
            "Content-Disposition",
            "attachment",
            filename=Path(attach).name,
        )
        return part

    def attaches_to_mimes(
        self, attaches: Union[list[Union[str, Path]], Union[str, Path]]
    ) -> list[MIMEBase]:
        if isinstance(attaches, (str, Path)):
            attaches = [attaches]
        mime_attaches = []
        for attach in attaches:
            mime_attach = self.attach_to_mime(attach)
            if mime_attach:
                mime_attaches.append(mime_attach)
        return mime_attaches

    def content_to_mime_msg(self, content: EmailContentType) -> str:
        msg = MIMEMultipart()

        # from, to, subject
        msg["From"] = self.username
        msg["To"] = self.mail_list_to_str(content["to"])
        msg["Subject"] = content["subject"]

        # cc, bcc
        if content.get("cc"):
            msg["Cc"] = self.mail_list_to_str(content["cc"])

        if content.get("bcc"):
            msg["Bcc"] = self.mail_list_to_str(content["bcc"])

        # body
        mime_type = content.get("mime_type", "html")
        mime_body = MIMEText(content["body"], mime_type)
        msg.attach(mime_body)

        # attachments
        mime_attaches = self.attaches_to_mimes(content.get("attachments", []))
        for mime_attach in mime_attaches:
            msg.attach(mime_attach)

        return msg.as_string()

    def send(self, content: EmailContentType) -> dict:
        logger.note(f"> Sending email:")
        logger.mesg(dict_to_str(content), indent=2)
        msg = self.content_to_mime_msg(content)
        if self.confirm_before_send:
            confirm_input("send", op_name="email send", max_retries=3)
        res = self.smtp.sendmail(self.username, content["to"], msg)
        if res:
            logger.mesg(dict_to_str(res), indent=2)
        else:
            logger.okay(f"+ Successfully sent to: {logstr.file(content['to'])}")
        return res

    def __del__(self):
        if self.smtp:
            self.disconnect()
        logger.exit_quiet(not self.verbose)
